@@ -165,7 +165,11 @@ def test_from_docstring_2():
             """Foo.bar(self, int ncall_me =10000, [resume=True, int nsplit=1])"""
             pass
 
+        def baz(self, *args):
+            """Foo.baz(self: Foo, ncall_me: int =10000, arg: np.ndarray = [])"""
+
     assert describe(Foo().bar) == ["ncall_me", "resume", "nsplit"]
+    assert describe(Foo().baz) == ["ncall_me", "arg"]
 
 
 def test_from_docstring_3():
@@ -196,8 +200,7 @@ def test_with_type_hints():
         d: Annotated[float, Ge(1)],
         e: Annotated[float, Le(2)],
         f: Annotated[float, Interval(gt=2, lt=3)],
-    ):
-        ...
+    ): ...
 
     r = describe(foo, annotations=True)
     assert r == {
@@ -211,8 +214,7 @@ def test_with_type_hints():
     }
 
     class Foo:
-        def __call__(self, x: NDArray, a: Annotated[float, Gt(0), Lt(1)], b: float):
-            ...
+        def __call__(self, x: NDArray, a: Annotated[float, Gt(0), Lt(1)], b: float): ...
 
     r = describe(Foo.__call__, annotations=True)
     assert r == {"self": None, "x": None, "a": (0, 1), "b": None}
@@ -231,8 +233,7 @@ def test_with_pydantic_types():
         c: float,
         d: Annotated[float, tp.annotated_types.Gt(1)],
         e: Annotated[float, tp.annotated_types.Interval(gt=0, lt=1)],
-    ):
-        ...
+    ): ...
 
     r = describe(foo, annotations=True)
     assert r == {
@@ -253,8 +254,7 @@ def test_with_annotated_types():
         a: float,
         b: Annotated[float, tp.Gt(1)],
         c: Annotated[float, tp.Interval(gt=0, lt=1)],
-    ):
-        ...
+    ): ...
 
     r = describe(foo, annotations=True)
     assert r == {
