@@ -4,7 +4,7 @@ import numpy as np
 from numpy.testing import assert_allclose, assert_equal
 from iminuit import Minuit
 from iminuit.util import Param, make_func_code
-from iminuit.warnings import IMinuitWarning
+from iminuit.warnings import IMinuitWarning, ErrordefAlreadySetWarning
 from iminuit.typing import Annotated
 from pytest import approx
 from argparse import Namespace
@@ -1693,3 +1693,37 @@ def test_annotated_cost_function():
     assert m2.limits["y"] == (0.1, 1.0)
     m.migrad()
     assert_allclose(m.values, (0, 0.1), atol=1e-2)
+
+
+def test_enforced_grad():
+    def cost(a, b):
+        return a**2 + b**2
+
+    with pytest.raises(ValueError):
+        Minuit(cost, 0, 0, grad=True)
+
+
+def test_bad_grad():
+    def cost(a, b):
+        return a**2 + b**2
+
+    with pytest.raises(ValueError, match="provided gradient is not a CostGradient"):
+        Minuit(cost, 0, 0, grad="foo")
+
+
+def test_errordef_already_set_warning():
+    def cost(a, b):
+        return a**2 + b**2
+
+    cost.errordef = 1
+
+    m = Minuit(cost, 0, 0)
+    m.hesse()
+    assert_allclose(m.errors, [1, 1])
+
+    with pytest.warns(ErrordefAlreadySetWarning):
+        m.errordef = 4
+
+    # check that cost.errordef value is still overridden
+    m.hesse()
+    assert_allclose(m.errors, [2, 2])
